"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.recursive = recursive;
exports.matchDependencies = matchDependencies;
exports.createMatcher = createMatcher;
exports.parseUpdateParam = parseUpdateParam;
exports.makeIgnorePatterns = makeIgnorePatterns;
const fs_1 = require("fs");
const path_1 = __importDefault(require("path"));
const cli_utils_1 = require("@pnpm/cli-utils");
const config_1 = require("@pnpm/config");
const error_1 = require("@pnpm/error");
const get_context_1 = require("@pnpm/get-context");
const logger_1 = require("@pnpm/logger");
const manifest_utils_1 = require("@pnpm/manifest-utils");
const matcher_1 = require("@pnpm/matcher");
const plugin_commands_rebuild_1 = require("@pnpm/plugin-commands-rebuild");
const pnpmfile_1 = require("@pnpm/pnpmfile");
const sort_packages_1 = require("@pnpm/sort-packages");
const store_connection_manager_1 = require("@pnpm/store-connection-manager");
const core_1 = require("@pnpm/core");
const is_subdir_1 = __importDefault(require("is-subdir"));
const mem_1 = __importDefault(require("mem"));
const p_filter_1 = __importDefault(require("p-filter"));
const p_limit_1 = __importDefault(require("p-limit"));
const updateWorkspaceDependencies_1 = require("./updateWorkspaceDependencies");
const getSaveType_1 = require("./getSaveType");
const getPinnedVersion_1 = require("./getPinnedVersion");
const errors_1 = require("./errors");
async function recursive(allProjects, params, opts, cmdFullName) {
    if (allProjects.length === 0) {
        // It might make sense to throw an exception in this case
        return false;
    }
    const pkgs = Object.values(opts.selectedProjectsGraph).map((wsPkg) => wsPkg.package);
    if (pkgs.length === 0) {
        return false;
    }
    const manifestsByPath = getManifestsByPath(allProjects);
    const throwOnFail = cli_utils_1.throwOnCommandFail.bind(null, `pnpm recursive ${cmdFullName}`);
    const store = opts.storeControllerAndDir ?? await (0, store_connection_manager_1.createOrConnectStoreController)(opts);
    const workspacePackages = (0, get_context_1.arrayOfWorkspacePackagesToMap)(allProjects);
    const targetDependenciesField = (0, getSaveType_1.getSaveType)(opts);
    const rootManifestDir = (opts.lockfileDir ?? opts.dir);
    const installOpts = Object.assign(opts, {
        ...(0, config_1.getOptionsFromRootManifest)(rootManifestDir, manifestsByPath[rootManifestDir]?.manifest ?? {}),
        allProjects: getAllProjects(manifestsByPath, opts.allProjectsGraph, opts.sort),
        linkWorkspacePackagesDepth: opts.linkWorkspacePackages === 'deep' ? Infinity : opts.linkWorkspacePackages ? 0 : -1,
        ownLifecycleHooksStdio: 'pipe',
        peer: opts.savePeer,
        pruneLockfileImporters: opts.pruneLockfileImporters ??
            (((opts.ignoredPackages == null) || opts.ignoredPackages.size === 0) &&
                pkgs.length === allProjects.length),
        storeController: store.ctrl,
        storeDir: store.dir,
        targetDependenciesField,
        workspacePackages,
        forceHoistPattern: typeof opts.rawLocalConfig?.['hoist-pattern'] !== 'undefined' || typeof opts.rawLocalConfig?.['hoist'] !== 'undefined',
        forceShamefullyHoist: typeof opts.rawLocalConfig?.['shamefully-hoist'] !== 'undefined',
    });
    const result = {};
    const memReadLocalConfig = (0, mem_1.default)(config_1.readLocalConfig);
    const updateToLatest = opts.update && opts.latest;
    const includeDirect = opts.includeDirect ?? {
        dependencies: true,
        devDependencies: true,
        optionalDependencies: true,
    };
    let updateMatch;
    if (cmdFullName === 'update') {
        if (params.length === 0) {
            const ignoreDeps = manifestsByPath[opts.workspaceDir]?.manifest?.pnpm?.updateConfig?.ignoreDependencies;
            if (ignoreDeps?.length) {
                params = makeIgnorePatterns(ignoreDeps);
            }
        }
        updateMatch = params.length ? createMatcher(params) : null;
    }
    else {
        updateMatch = null;
    }
    // For a workspace with shared lockfile
    if (opts.lockfileDir && ['add', 'install', 'remove', 'update', 'import'].includes(cmdFullName)) {
        let importers = getImporters(opts);
        const calculatedRepositoryRoot = await fs_1.promises.realpath(calculateRepositoryRoot(opts.workspaceDir, importers.map(x => x.rootDir)));
        const isFromWorkspace = is_subdir_1.default.bind(null, calculatedRepositoryRoot);
        importers = await (0, p_filter_1.default)(importers, async ({ rootDirRealPath }) => isFromWorkspace(rootDirRealPath));
        if (importers.length === 0)
            return true;
        let mutation;
        switch (cmdFullName) {
            case 'remove':
                mutation = 'uninstallSome';
                break;
            case 'import':
                mutation = 'install';
                break;
            default:
                mutation = (params.length === 0 && !updateToLatest ? 'install' : 'installSome');
                break;
        }
        const mutatedImporters = [];
        await Promise.all(importers.map(async ({ rootDir }) => {
            const localConfig = await memReadLocalConfig(rootDir);
            const modulesDir = localConfig.modulesDir ?? opts.modulesDir;
            const { manifest } = manifestsByPath[rootDir];
            let currentInput = [...params];
            if (updateMatch != null) {
                currentInput = matchDependencies(updateMatch, manifest, includeDirect);
                if ((currentInput.length === 0) && (typeof opts.depth === 'undefined' || opts.depth <= 0)) {
                    installOpts.pruneLockfileImporters = false;
                    return;
                }
            }
            if (updateToLatest && (!params || (params.length === 0))) {
                currentInput = Object.keys((0, manifest_utils_1.filterDependenciesByType)(manifest, includeDirect));
            }
            if (opts.workspace) {
                if (!currentInput || (currentInput.length === 0)) {
                    currentInput = (0, updateWorkspaceDependencies_1.updateToWorkspacePackagesFromManifest)(manifest, includeDirect, workspacePackages);
                }
                else {
                    currentInput = (0, updateWorkspaceDependencies_1.createWorkspaceSpecs)(currentInput, workspacePackages);
                }
            }
            switch (mutation) {
                case 'uninstallSome':
                    mutatedImporters.push({
                        dependencyNames: currentInput,
                        modulesDir,
                        mutation,
                        rootDir,
                        targetDependenciesField,
                    });
                    return;
                case 'installSome':
                    mutatedImporters.push({
                        allowNew: cmdFullName === 'install' || cmdFullName === 'add',
                        dependencySelectors: currentInput,
                        modulesDir,
                        mutation,
                        peer: opts.savePeer,
                        pinnedVersion: (0, getPinnedVersion_1.getPinnedVersion)({
                            saveExact: typeof localConfig.saveExact === 'boolean' ? localConfig.saveExact : opts.saveExact,
                            savePrefix: typeof localConfig.savePrefix === 'string' ? localConfig.savePrefix : opts.savePrefix,
                        }),
                        rootDir,
                        targetDependenciesField,
                        update: opts.update,
                        updateMatching: opts.updateMatching,
                        updatePackageManifest: opts.updatePackageManifest,
                        updateToLatest: opts.latest,
                    });
                    return;
                case 'install':
                    mutatedImporters.push({
                        modulesDir,
                        mutation,
                        pruneDirectDependencies: opts.pruneDirectDependencies,
                        rootDir,
                        update: opts.update,
                        updateMatching: opts.updateMatching,
                        updatePackageManifest: opts.updatePackageManifest,
                        updateToLatest: opts.latest,
                    });
            }
        }));
        if (!opts.selectedProjectsGraph[opts.workspaceDir] && manifestsByPath[opts.workspaceDir] != null) {
            mutatedImporters.push({
                mutation: 'install',
                rootDir: opts.workspaceDir,
            });
        }
        if ((mutatedImporters.length === 0) && cmdFullName === 'update' && opts.depth === 0) {
            throw new error_1.PnpmError('NO_PACKAGE_IN_DEPENDENCIES', 'None of the specified packages were found in the dependencies of any of the projects.');
        }
        const { updatedProjects: mutatedPkgs, ignoredBuilds } = await (0, core_1.mutateModules)(mutatedImporters, {
            ...installOpts,
            storeController: store.ctrl,
        });
        if (opts.save !== false) {
            await Promise.all(mutatedPkgs
                .map(async ({ originalManifest, manifest, rootDir }) => {
                return manifestsByPath[rootDir].writeProjectManifest(originalManifest ?? manifest);
            }));
        }
        if (opts.strictDepBuilds && ignoredBuilds?.length) {
            throw new errors_1.IgnoredBuildsError(ignoredBuilds);
        }
        return true;
    }
    const pkgPaths = Object.keys(opts.selectedProjectsGraph).sort();
    const limitInstallation = (0, p_limit_1.default)((0, config_1.getWorkspaceConcurrency)(opts.workspaceConcurrency));
    await Promise.all(pkgPaths.map(async (rootDir) => limitInstallation(async () => {
        const hooks = opts.ignorePnpmfile
            ? {}
            : (() => {
                const pnpmfileHooks = (0, pnpmfile_1.requireHooks)(rootDir, opts);
                return {
                    ...opts.hooks,
                    ...pnpmfileHooks,
                    afterAllResolved: [...(pnpmfileHooks.afterAllResolved ?? []), ...(opts.hooks?.afterAllResolved ?? [])],
                    readPackage: [...(pnpmfileHooks.readPackage ?? []), ...(opts.hooks?.readPackage ?? [])],
                };
            })();
        try {
            if (opts.ignoredPackages?.has(rootDir)) {
                return;
            }
            result[rootDir] = { status: 'running' };
            const { manifest, writeProjectManifest } = manifestsByPath[rootDir];
            let currentInput = [...params];
            if (updateMatch != null) {
                currentInput = matchDependencies(updateMatch, manifest, includeDirect);
                if (currentInput.length === 0)
                    return;
            }
            if (updateToLatest && (!params || (params.length === 0))) {
                currentInput = Object.keys((0, manifest_utils_1.filterDependenciesByType)(manifest, includeDirect));
            }
            if (opts.workspace) {
                if (!currentInput || (currentInput.length === 0)) {
                    currentInput = (0, updateWorkspaceDependencies_1.updateToWorkspacePackagesFromManifest)(manifest, includeDirect, workspacePackages);
                }
                else {
                    currentInput = (0, updateWorkspaceDependencies_1.createWorkspaceSpecs)(currentInput, workspacePackages);
                }
            }
            let action;
            switch (cmdFullName) {
                case 'remove':
                    action = async (manifest, opts) => {
                        const mutationResult = await (0, core_1.mutateModules)([
                            {
                                dependencyNames: currentInput,
                                mutation: 'uninstallSome',
                                rootDir,
                            },
                        ], opts);
                        return { updatedManifest: mutationResult.updatedProjects[0].manifest, ignoredBuilds: mutationResult.ignoredBuilds };
                    };
                    break;
                default:
                    action = currentInput.length === 0
                        ? core_1.install
                        : async (manifest, opts) => (0, core_1.addDependenciesToPackage)(manifest, currentInput, opts);
                    break;
            }
            const localConfig = await memReadLocalConfig(rootDir);
            const { updatedManifest: newManifest, ignoredBuilds } = await action(manifest, {
                ...installOpts,
                ...localConfig,
                ...(0, config_1.getOptionsFromRootManifest)(rootDir, manifest),
                ...opts.allProjectsGraph[rootDir]?.package,
                bin: path_1.default.join(rootDir, 'node_modules', '.bin'),
                dir: rootDir,
                hooks,
                ignoreScripts: true,
                pinnedVersion: (0, getPinnedVersion_1.getPinnedVersion)({
                    saveExact: typeof localConfig.saveExact === 'boolean' ? localConfig.saveExact : opts.saveExact,
                    savePrefix: typeof localConfig.savePrefix === 'string' ? localConfig.savePrefix : opts.savePrefix,
                }),
                rawConfig: {
                    ...installOpts.rawConfig,
                    ...localConfig,
                },
                storeController: store.ctrl,
            });
            if (opts.save !== false) {
                await writeProjectManifest(newManifest);
            }
            if (opts.strictDepBuilds && ignoredBuilds?.length) {
                throw new errors_1.IgnoredBuildsError(ignoredBuilds);
            }
            result[rootDir].status = 'passed';
        }
        catch (err) { // eslint-disable-line
            logger_1.logger.info(err);
            if (!opts.bail) {
                result[rootDir] = {
                    status: 'failure',
                    error: err,
                    message: err.message,
                    prefix: rootDir,
                };
                return;
            }
            err['prefix'] = rootDir;
            throw err;
        }
    })));
    if (!opts.lockfileOnly && !opts.ignoreScripts && (cmdFullName === 'add' ||
        cmdFullName === 'install' ||
        cmdFullName === 'update')) {
        await plugin_commands_rebuild_1.rebuild.handler({
            ...opts,
            pending: opts.pending === true,
            skipIfHasSideEffectsCache: true,
        }, []);
    }
    throwOnFail(result);
    if (!Object.values(result).filter(({ status }) => status === 'passed').length && cmdFullName === 'update' && opts.depth === 0) {
        throw new error_1.PnpmError('NO_PACKAGE_IN_DEPENDENCIES', 'None of the specified packages were found in the dependencies of any of the projects.');
    }
    return true;
}
function calculateRepositoryRoot(workspaceDir, projectDirs) {
    // assume repo root is workspace dir
    let relativeRepoRoot = '.';
    for (const rootDir of projectDirs) {
        const relativePartRegExp = new RegExp(`^(\\.\\.\\${path_1.default.sep})+`);
        const relativePartMatch = relativePartRegExp.exec(path_1.default.relative(workspaceDir, rootDir));
        if (relativePartMatch != null) {
            const relativePart = relativePartMatch[0];
            if (relativePart.length > relativeRepoRoot.length) {
                relativeRepoRoot = relativePart;
            }
        }
    }
    return path_1.default.resolve(workspaceDir, relativeRepoRoot);
}
function matchDependencies(match, manifest, include) {
    const deps = Object.keys((0, manifest_utils_1.filterDependenciesByType)(manifest, include));
    const matchedDeps = [];
    for (const dep of deps) {
        const spec = match(dep);
        if (spec === null)
            continue;
        matchedDeps.push(spec ? `${dep}@${spec}` : dep);
    }
    return matchedDeps;
}
function createMatcher(params) {
    const patterns = [];
    const specs = [];
    for (const param of params) {
        const { pattern, versionSpec } = parseUpdateParam(param);
        patterns.push(pattern);
        specs.push(versionSpec ?? '');
    }
    const matcher = (0, matcher_1.createMatcherWithIndex)(patterns);
    return (depName) => {
        const index = matcher(depName);
        if (index === -1)
            return null;
        return specs[index];
    };
}
function parseUpdateParam(param) {
    const atIndex = param.indexOf('@', param[0] === '!' ? 2 : 1);
    if (atIndex === -1) {
        return {
            pattern: param,
            versionSpec: undefined,
        };
    }
    return {
        pattern: param.slice(0, atIndex),
        versionSpec: param.slice(atIndex + 1),
    };
}
function makeIgnorePatterns(ignoredDependencies) {
    return ignoredDependencies.map(depName => `!${depName}`);
}
function getAllProjects(manifestsByPath, allProjectsGraph, sort) {
    const chunks = sort !== false
        ? (0, sort_packages_1.sortPackages)(allProjectsGraph)
        : [Object.keys(allProjectsGraph).sort()];
    return chunks.map((prefixes, buildIndex) => prefixes.map((rootDir) => {
        const { rootDirRealPath, modulesDir } = allProjectsGraph[rootDir].package;
        return {
            buildIndex,
            manifest: manifestsByPath[rootDir].manifest,
            rootDir,
            rootDirRealPath,
            modulesDir,
        };
    })).flat();
}
function getManifestsByPath(projects) {
    const manifestsByPath = {};
    for (const { rootDir, manifest, writeProjectManifest } of projects) {
        manifestsByPath[rootDir] = { manifest, writeProjectManifest };
    }
    return manifestsByPath;
}
function getImporters(opts) {
    let rootDirs = Object.keys(opts.selectedProjectsGraph);
    if (opts.ignoredPackages != null) {
        rootDirs = rootDirs.filter((rootDir) => !opts.ignoredPackages.has(rootDir));
    }
    return rootDirs.map((rootDir) => ({ rootDir, rootDirRealPath: opts.selectedProjectsGraph[rootDir].package.rootDirRealPath }));
}
//# sourceMappingURL=recursive.js.map